Require Import Vtac Program List Arith Even Div2 Relations String.
Require Import basis countable lang subs model basiclemmas.
Require Import Classical FunctionalExtensionality.

Open Scope string_scope.

Set Implicit Arguments.


Section wkmorph. (****************************************************************************)

Variables w1 w2 : world.

Definition morph := wst w1 -> wst w2 -> Prop.

Inductive is_wkmorph (m : morph) : Prop :=
  is_wkmorph_intro
    (MNAM : forall a, wnam w1 a <-> wnam w2 a)
    (MEXT : forall s1 s1' s2 s2' (EXT1: wext w1 s1' s1) (M12: m s1 s2)
      (M12': m s1' s2'), wext w2 s2' s2)
    (MEXTPUB : forall s1 s1' s2 s2' (EP : wextpub w1 s1' s1) (M12 : m s1 s2)
      (M12' : m s1' s2'), wextpub w2 s2' s2)
    (MVAL : forall s1 s2 g (M12 : m s1 s2) ty v1 v2,
      wval w1 () s1 g ty v1 v2 <-> wval w2 () s2 g ty v1 v2)
    (MH : forall s1 g1 h1 h2 (G1 : wglob w1 g1) (H12 : wheap w1 () s1 (g1 s1) h1 h2),
      exists s2, m s1 s2 /\ wheap w2 () s2 (g1 s1) h1 h2)
  .

Definition wmext (m' m : morph) : Prop :=
  forall s1 s2 s2' (M12 : m s1 s2) (M12' : m' s1 s2'), wext w2 s2' s2.
Definition wmextpub (m' m : morph) : Prop :=
  forall s1 s2 s2' (M12 : m s1 s2) (M12' : m' s1 s2'), wextpub w2 s2' s2.

End wkmorph. (********************************************************************************)

Hint Constructors is_wkmorph.

Section wkiso. (******************************************************************************)

Variables w1 w2 : world.

Definition morph_id w : morph w w :=
  fun s s' => s = s'.

Inductive morph_comp w1 w2 w3 (m2 : morph w2 w3) (m1 : morph w1 w2) s1 s3 : Prop :=
  morph_comp_intro s2 (M12 : m1 s1 s2) (M23 : m2 s2 s3).

Lemma morph_id_is_wkmorph : forall w, is_wkmorph (morph_id w).
Proof.
  intros; apply is_wkmorph_intro; try red; intros; auto;
    try inversion M12 as (E12); try rewrite <- E12;
    try inversion M12' as (E12'); try rewrite <- E12'; unfold morph_id; eauto.
Qed.

Structure iso w1 w2 := mkiso
  { iso1 : morph w1 w2
  ; iso2 : morph w2 w1 }.

Inductive is_wkiso w1 w2 (i : iso w1 w2) : Prop :=
  is_wkiso_intro
    (M1 : is_wkmorph (iso1 i))
    (M2 : is_wkmorph (iso2 i))
    (COMP12 : wmextpub (morph_comp (iso1 i) (iso2 i)) (morph_id _))
    (COMP21 : wmextpub (morph_comp (iso2 i) (iso1 i)) (morph_id _))
  .

End wkiso. (**********************************************************************************)

Hint Constructors is_wkiso.

Hint Unfold morph_id.
Hint Constructors morph_comp.

Section wglob. (******************************************************************************)

Variables w1 w2 : world.

Definition morphgl_cup (m : morph w1 w2) (r2 : wst w2 -> vrelf) s1 ty v1 v2 : Prop :=
  exists s2 s1', m s1' s2 /\ r2 s2 ty v1 v2 /\ wext w1 s1 s1'.
Hint Unfold morphgl_cup.

Inductive morphgl (m : morph w1 w2) (g2 : wst w2 -> vrelf) s1 
                  (ty: flextyp) (v1 v2: val) : Prop :=
| morphgl_base (REL: morphgl_cup m g2 s1 ty v1 v2)
| morphgl_step r (REL: wval w1 () s1 r ty v1 v2) (MON: r <3= morphgl m g2 s1).
Hint Constructors morphgl.

Lemma morphgl_eq_aux : forall (m : morph w1 w2) (M : is_wkmorph m) (g2 : wst w2 -> vrelf)
  (G2 : wglob w2 g2) s1 s2 (M12 : m s1 s2), g2 s2 = morphgl_cup m g2 s1.
Proof.
  intros; extensionality ty; extensionality v1; extensionality v2.
  apply prop_ext; split; intros IN.
    by exists s2 s1; eauto.
  destruct IN as (s2', (s1', (M12', (EXT, IN')))).
  destruct G2 as (_ & MON & _); eapply MON; eauto.
  by destruct M; eapply MEXT; eauto.
Qed.

Lemma morphgl_eq : forall (m : morph w1 w2) (M : is_wkmorph m) (g2 : wst w2 -> vrelf)
  (G2 : wglob w2 g2) s1 s2 (M12 : m s1 s2), morphgl m g2 s1 = g2 s2.
Proof.
  intros; extensionality ty; extensionality v1; extensionality v2.
  apply prop_ext; split; intro IN.
    induction IN; [eby erewrite <- morphgl_eq_aux in REL|].
    by des; eauto; destruct M; rdes G2; rhdes GSUP; eauto using wval_mon.
  by econstructor; try erewrite <- morphgl_eq_aux; eauto.
Qed.

Lemma morphgl_mon : forall (m : morph w1 w2) (g2 : wst w2 -> vrelf) (G2 : wglob w2 g2)
  s1 s1' (EXT : wext w1 s1' s1), morphgl m g2 s1 <3= morphgl m g2 s1'.
Proof.
  intros; induction PR; rdes REL; eauto 10 using wval_mon, wext_trans.
Qed.

Lemma morphgl_glob : forall (m : morph w1 w2) (M : is_wkmorph m) (g2 : wst w2 -> vrelf)
  (G2 : wglob w2 g2), wglob w1 (morphgl m g2).
Proof.
  intros; split; [| split]; unnw; intros.
  - by red; intros; rdes G2; rhdes GSUP; red in GCLO; induction V;
       [rdes REL | eapply wval_closed]; eauto.
  - by eapply morphgl_mon; eauto.
  - split; try split; unnw; intros; eauto using wval_mon.
    + destruct R; eauto using wval_mon.
      assert (G2':=G2); assert (M':=M); destruct M'; rdes G2'; rhdes GSUP; rdes REL.
      eapply GEREF in REL0; eauto.
      by rewrite <- (morphgl_eq M G2 s1') in REL0; eauto using wval_mon, morphgl_mon.
    + destruct R; eauto using wval_mon.
      assert (G2':=G2); assert (M':=M); destruct M'; rdes G2'; rhdes GSUP; rdes REL.
      eapply GENAM in REL0; eauto.
      by rewrite <- (morphgl_eq M G2 s1') in REL0; eauto using wval_mon, morphgl_mon.
Qed.

Definition morphgr_cup (m : morph w1 w2) (r1 : wst w1 -> vrelf) s1 s2 ty v1 v2 : Prop :=
  exists s2' s1', m s1' s2' /\ r1 s1' ty v1 v2 /\ wext w1 s1 s1' /\ wext w2 s2 s2'.
Hint Unfold morphgr_cup.

Inductive morphgr (m : morph w1 w2) (g1 : wst w1 -> vrelf) s1 s2 
                  (ty: flextyp) (v1 v2: val) : Prop :=
| morphgr_base (REL: morphgr_cup m g1 s1 s2 ty v1 v2)
| morphgr_step r (REL: wval w2 () s2 r ty v1 v2) (MON: r <3= morphgr m g1 s1 s2).
Hint Constructors morphgr.

Lemma morphgr_eq_aux : forall (m : morph w1 w2) (M : is_wkmorph m) (g1 : wst w1 -> vrelf)
  (G1 : wglob w1 g1) s1 s2 (M12 : m s1 s2), g1 s1 = morphgr_cup m g1 s1 s2.
Proof.
  intros; extensionality ty; extensionality v1; extensionality v2.
  apply prop_ext; split; intros IN.
    exists s2 s1; split; auto.
  destruct IN as (s1' & s2' & ? & ? & ? & ?).
  by destruct G1 as (_ & MON & _); eapply MON; eauto.
Qed.

Lemma morphgr_eq : forall (m : morph w1 w2) (M : is_wkmorph m) (g1 : wst w1 -> vrelf)
  (G1 : wglob w1 g1) s1 s2 (M12 : m s1 s2), morphgr m g1 s1 s2 = g1 s1.
Proof.
  intros; extensionality ty; extensionality v1; extensionality v2.
  apply prop_ext; split; intro IN.
    induction IN; [eby erewrite <- morphgr_eq_aux in REL|].
    by des; eauto; destruct M; rdes G1; rhdes GSUP; eauto using wval_mon.
  by econstructor; try erewrite <- morphgr_eq_aux; eauto.
Qed.

Lemma morphgr_mon : forall (m : morph w1 w2) (g1 : wst w1 -> vrelf) (G1 : wglob w1 g1)
    s1 s1' (EXT1 : wext w1 s1' s1) s2 s2' (EXT2 : wext w2 s2' s2),
  morphgr m g1 s1 s2 <3= morphgr m g1 s1' s2'.
Proof.
  intros; induction PR; rdes REL; eauto 10 using wval_mon, wext_trans.
Qed.

Lemma morphgr_glob : forall (m : morph w1 w2) (M : is_wkmorph m) (g1 : wst w1 -> vrelf)
  (G1 : wglob w1 g1) s1, wglob w2 (morphgr m g1 s1).
Proof.
  intros; split; [| split]; unnw; intros.
  - by red; intros; rdes G1; rhdes GSUP; red in GCLO; induction V;
       [rdes REL | eapply wval_closed]; eauto.
  - by eapply morphgr_mon; eauto.
  - split; try split; unnw; intros; eauto using wval_mon.
    + destruct R; eauto using wval_mon.
      assert (G1':=G1); assert (M':=M); destruct M'; rdes G1'; rhdes GSUP; rdes REL.
      eapply GEREF in REL0; eauto.
      by rewrite <- (morphgr_eq M G1 s1' s2') in REL0; eauto using wval_mon, morphgr_mon.
    + destruct R; eauto using wval_mon.
      assert (G1':=G1); assert (M':=M); destruct M'; rdes G1'; rhdes GSUP; rdes REL.
      eapply GENAM in REL0; eauto.
      by rewrite <- (morphgr_eq M G1 s1' s2') in REL0; eauto using wval_mon, morphgr_mon.
Qed.

End wglob. (**********************************************************************************)

Hint Unfold morphgl_cup morphgr_cup.
Hint Constructors morphgl morphgr.

Section category_laws.

Lemma wkmorph_composable: forall w1 w2 w3 (m1 : morph w1 w2) (m2 : morph w2 w3)
    (WK1: is_wkmorph m1)
    (WK2: is_wkmorph m2),
  is_wkmorph (morph_comp m2 m1).
Proof.
  intros; destruct (id WK1), (id WK2); hdes; econstructor; 
   intros; try destruct M12; try destruct M12'; eauto 10.

  specialize (MH _ _ _ _ G1 H12); des.
  exploit MH0; eauto using morphgr_glob.
    eapply wheap_mon; intros; eauto.
    by erewrite morphgr_eq; eauto.
  by intros; des; rewrite morphgr_eq in x1; eauto.
Qed.

Lemma morph_identity1: forall w1 w2 (m: morph w1 w2),
  morph_comp (morph_id _) m = m.
Proof. 
  intros; extensionality s1; extensionality s2.
  eapply prop_ext; split; intros.
  - by destruct H;  red in M23; subst.
  - by eauto.
Qed.

Lemma morph_identity2: forall w1 w2 (m: morph w1 w2),
  morph_comp m (morph_id _) = m.
Proof. 
  intros; extensionality s1; extensionality s2.
  eapply prop_ext; split; intros.
  - by destruct H; red in M12; subst.
  - by eauto.
Qed.

Lemma morph_assoc: forall w1 w2 w3 w4 
    (m1: morph w1 w2) (m2: morph w2 w3) (m3: morph w3 w4),
  morph_comp (morph_comp m3 m2) m1 = morph_comp m3 (morph_comp m2 m1).
Proof.
  intros; extensionality s1; extensionality s2.
  eapply prop_ext; split; intros.
  - by destruct H; destruct M23; eauto.
  - by destruct H; destruct M12; eauto.
Qed.

End category_laws.

Section equiv. (******************************************************************************)

Variables w1 w2 : world.

Definition gc (m : morph w1 w2) (S1 : wst w1 -> Prop) s1 : Prop :=
 S1 s1 /\ exists s2, m s1 s2.

Definition cupcap (i : iso w1 w2) s2 s2' (f : wst w1 -> wst w1 -> erel) : erel :=
  fun ty e1 e2 =>
    exists s1, gc (iso1 i) (iso2 i s2) s1 /\
      forall s1' (GC : gc (iso1 i) (iso2 i s2') s1'),
        f s1 s1' ty e1 e2.

Lemma wexp_iso_aux : forall (i : iso w1 w2) (LI : is_wkiso i) g2 (g2OK: wglob w2 g2) s2 s2',
  cupcap i s2 s2' (wexp w1 (morphgl (iso1 i) g2)) <3= wexp w2 g2 s2 s2'.
Proof.
  cofix.
  intros until s2'; intros ty v1 v2 IN; repeat intro.
  destruct IN as (s1 & IN1 & IN2); edestructs LI M2; apply MH in HREL; auto.
  destruct HREL as (s1', (I2, HREL)); assert (HREL' := HREL).
  erewrite <- morphgr_eq in HREL; eauto.
  edestruct M1; apply MH0 in HREL; [| by eapply morphgr_glob].
  destruct HREL as (s2'' & I1'2'' & HREL).
  edestruct g2OK as (_ & MON & _).
  eapply wheap_mon with (r' := g2 s2'') in HREL'; unnw; eauto 10 using wextpub_in_wext.
  erewrite <- morphgl_eq in HREL'; eauto.
  assert (IN : wexp_core w1 (morphgl (iso1 i) g2)
               (morphgl (iso1 i) g2) s1 s1' ty h hf hh v1 h' hf' hh' v2).
    by eapply IN2; unfold gc; eauto. 
  destruct IN.

  Case "Diverge".
  by apply wexp_Diverge.

  Case "Terminate".
  apply MH0 in HREL0; [| by apply morphgl_glob].
  destruct HREL0 as (s2''' & I1'' & HREL0).
  erewrite morphgl_eq in HREL0; eauto.
  eapply wexp_Terminate with (s' := s2'''); eauto.
  - eapply wext_trans.
      by eapply MEXT0; [apply EXT | apply I1'2'' | ].
    apply wextpub_in_wext; eapply COMP12 with s2'. 
      by red; auto.
    by eapply morph_comp_intro; [apply I2 | apply I1'2''].
  - destruct IN1 as (B & s2''0 & A).
    eapply wextpub_trans.
      by eapply MEXTPUB0; [apply EXTP | apply A | ].
    eapply COMP12 with s2.
      by red; auto.
    by eapply morph_comp_intro; [apply B | apply A ].
  - by erewrite <- morphgl_eq; try apply VREL; try apply I1''.
  
  Case "Call".
  apply MH0 in HREL0; [| by apply morphgl_glob].
  destruct HREL0 as (s2''' & I1'' & HREL0).
  erewrite morphgl_eq in HREL0; eauto.
  unfold id in EREL; erewrite morphgl_eq in EREL; eauto.
  eapply wexp_Call with (s' := s2'''); eauto.
  - eapply wext_trans.
      by eapply MEXT0; [apply EXT | apply I1'2'' | apply I1''].
    apply wextpub_in_wext; eapply COMP12 with s2'. 
      by red; auto.
    by eapply morph_comp_intro; [apply I2 | apply I1'2''].
  - repeat intro; eapply wexp_iso_aux; eauto.
    red; exists s1; split; auto; intros; apply MH in HREL0; eauto.
    destruct HREL0 as (s1'' & I & _); eapply KREL; eauto using morphgl_glob.
    + destruct GC; eapply wextpub_trans.
        by eapply MEXTPUB; [by apply EXTP | by apply I | by auto].
      eapply COMP21 with s'; [by red; auto |].
      by eapply morph_comp_intro; [by apply I1'' | by auto].
    + by intros; induction PR; rdes REL; eauto 10.
    + destruct GC as (I2' & st2' & I1'''); rewrite morphgl_eq with (s2 := st2'); eauto.
      eapply (vclos_mon _ VREL); eauto; auto.
      destruct g'OK as (_ & MON' & _); repeat intro; eapply MON'; eauto.
      by apply wextpub_in_wext; eapply COMP12 with s''; eauto.
Qed.

Lemma wexp_iso : forall g2 (g2OK: wglob w2 g2) s2 ty e1 e2
  (i : iso w1 w2)
  (LI : is_wkiso i)
  (IN : forall s1 (M12 : gc (iso1 i) (iso2 i s2) s1),
          wexp w1 (morphgl (iso1 i) g2) s1 s1 ty e1 e2),
  wexp w2 g2 s2 s2 ty e1 e2.
Proof.
  do 2 red; intros.
  edestructs LI M2; edestruct MH as (s1 & ? & B); eauto.
  eapply wheap_mon in B.
  2 : by erewrite <- morphgr_eq; eauto.
  edestruct M1; edestruct MH0 as (s2' & ? & D); eauto.
    by apply morphgr_glob.
  eapply wexp_iso_aux; eauto.
  exists s1; split; try red; eauto; intros.
  eapply wexp_mon; eauto.
  by eapply MEXTPUB; [apply wextpub_refl | eauto | destruct GC as (I, _); apply I].
Qed.

Lemma wstuck_beta : forall (w : world) s r r' ty e1 e2
    (ST : wstuck (wval w () s r) r' ty e1 e2),
  (exists e1', forall h : heap, reduce h e1 h e1') /\ 
  (exists e2', forall h : heap, reduce h e2 h e2').
Proof.
  intros; destruct ST; rewrite EQ1, EQ2; subst.
  - assert (X := wval_arrow_wf); hdesH X.
    by split; [eapply FV1|eapply FV2]; eauto.
  - assert (X := wval_forall_wf); hdesH X.
    by split; [eapply FV1|eapply FV2]; eauto.
Qed.

Lemma winhabited_iso : forall (i : iso w1 w2) (LI : is_wkiso i) (INH: winhabited w1), 
  winhabited w2.
Proof.
  repeat intro; destruct (id LI); destruct (id M1).
  edestruct INH; eauto using morphgl_glob.
  edestruct MH; eauto using morphgl_glob. 
  by exists x0; erewrite <- morphgl_eq; try apply H0; auto.
Qed.

Lemma wconsistent_iso : forall (i : iso w1 w2) (LI : is_wkiso i) (CS: wconsistent w1), 
  wconsistent w2.
Proof.
  repeat intro; destruct (id LI); destruct (id M1); destruct (id M2).
  assert (gOK0:=gOK); red in gOK0; des.
  edestruct wstuck_beta as ((e1' & A) & (e2' & B)); unnw.
    by apply APP.
  exists e1' e2'; split; try split; try apply A; try apply B.
  eapply wexp_iso; auto.
    by apply LI.
  intros; destruct M12 as (I2, (s2, I1)).
  edestruct CS as (? & ? & A1 & A2 & A3).
      by apply morphgl_glob; eauto.
    2 : assert (E1 : x = e1'); [eapply reduce_det |]; [ by apply A1 | by apply A |  ].
    2 : assert (E2 : x0 = e2'); [eapply reduce_det |]; [ by apply A2 | by apply B |  ].
    2 : rewrite <- E1, <- E2; apply A3.
    destruct APP.

    Case "application".
    eapply wstuck_App; [by apply EQ1 | by apply EQ2 | | ].
      eapply MVAL0; [by apply I2|].
      erewrite morphgl_eq; try apply I1; eauto.
      by eauto 10 using wval_mon, wextpub_in_wext.
    by erewrite morphgl_eq; try apply I1; eauto 10 using vclos_mon, wextpub_in_wext.

    Case "instantiation".
    eapply wstuck_Inst; [by apply EQ1 | by apply EQ2 | by apply EQ3 | done | ].
    eapply MVAL0; [by apply I2|].
      erewrite morphgl_eq; try apply I1; eauto.
      by eauto 10 using wval_mon, wextpub_in_wext.

  Existential 1:=hemp. Existential 1:=hemp.
Qed.

Theorem weq_iso : forall D G ty e1 e2
   (i : iso w1 w2)
   (LI : is_wkiso i)
   (EQ1 : weq D G w1 ty e1 e2),
   weq D G w2 ty e1 e2.
Proof.
  intros; rdes EQ1; destructs LI M1 M2; split; [|split]; try red; intros; eauto.

  Case "inhabited".
  eby eapply winhabited_iso.

  Case "consistent".
  eby eapply wconsistent_iso.

  Case "wequiv".
  destruct EQ; constructor; eauto.
  intros; assert (gOK0:=gOK); red in gOK0; des.
  eapply wexp_iso; eauto.
  intros; destruct M12 as (I2, (s2, I1)); apply WEXP; eauto using morphgl_glob; intros.
  erewrite morphgl_eq; try apply I1; eauto 10 using vclos_mon, wextpub_in_wext.
Qed.

End equiv. (******************************************************************************)


Section sym. (***************************************************************************)

Lemma is_wkiso_sym : forall (w1 w2 : world) (i : iso w1 w2) (LI : is_wkiso i),
  is_wkiso (mkiso (iso2 i) (iso1 i)).
Proof.
  intros; destruct LI; destruct M1, M2.
  by apply is_wkiso_intro; eauto; apply is_wkmorph_intro; eauto. 
Qed.

End sym. (*******************************************************************************)
